@@ -14,46 +14,31 @@ GNU Lesser General Public License for more details.
 You should have received a copy of the GNU Lesser General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.
 '''
-# suppress all warnings.
-import textwrap
-import rich
-import shlex
-import sys
-import os
-import re
-import difflib
-import argparse
-import concurrent.futures
-from prompt_toolkit.completion import Completer, Completion
-from prompt_toolkit import PromptSession
+from debgpt import version
+from pygments import highlight
+from pygments.formatters import TerminalFormatter
+from pygments.lexers import DiffLexer
+from rich.markup import escape
 from rich.panel import Panel
 from rich.rule import Rule
-from rich.markup import escape
-from rich.progress import track
-from prompt_toolkit.styles import Style
-from pygments import highlight
-from pygments.lexers import DiffLexer
-from pygments.formatters import TerminalFormatter
-from typing import List, Optional
-import warnings
+from typing import Optional
+import difflib
 import functools as ft
-from . import defaults
-from . import reader
-from . import frontend
-from . import configurator
-from . import arguments
-from debgpt import version
 import os
 import sys
-from rich.panel import Panel
 import tempfile
 import textwrap
-from . import frontend
-from . import reader
+import warnings
+
+from . import arguments
+from . import configurator
 from . import defaults
-from . import vectordb
-from . import replay
+from . import frontend
 from . import mapreduce
+from . import reader
+from . import replay
+from . import vectordb
+from . import cache
 
 warnings.filterwarnings("ignore")
 
@@ -151,34 +136,34 @@ def subcmd_git_commit(ag) -> None:
                 width=80))
         commit_message += '\n'
         commit_message += '\n'.join(
-            textwrap.wrap(f"\n\nThe real prompt is: {ag.ask}.", width=80))
+            textwrap.wrap(f"\n\nThe real prompt is: {repr(ag.ask)}", width=80))
         commit_message += '\n'
         commit_message += '\n'.join(
-            textwrap.wrap(f"\n\nFrontend used: {ag.frontend}.", width=80))
+            textwrap.wrap(f"\n\nFrontend used: {repr(ag.frontend)}", width=80))
         commit_message += '\n'
         if ag.frontend == "openai":
             commit_message += '\n'.join(
-                textwrap.wrap(f"\n\nOpenAI model: {ag.openai_model}.",
+                textwrap.wrap(f"\n\nOpenAI model: {repr(ag.openai_model)}",
                               width=80))
         elif ag.frontend == "google":
             commit_message += '\n'.join(
-                textwrap.wrap(f"\n\nGoogle model: {ag.google_model}.",
+                textwrap.wrap(f"\n\nGoogle model: {repr(ag.google_model)}",
                               width=80))
         elif ag.frontend == "anthropic":
             commit_message += '\n'.join(
-                textwrap.wrap(f"\n\nAnthropic model: {ag.anthropic_model}.",
+                textwrap.wrap(f"\n\nAnthropic model: {repr(ag.anthropic_model)}",
                               width=80))
         elif ag.frontend == "ollama":
             commit_message += '\n'.join(
-                textwrap.wrap(f"\n\nOllama model: {ag.ollama_model}.",
+                textwrap.wrap(f"\n\nOllama model: {repr(ag.ollama_model)}",
                               width=80))
         elif ag.frontend == "llamafile":
             commit_message += '\n'.join(
-                textwrap.wrap(f"\n\nLlamafile model: {ag.llamafile_model}.",
+                textwrap.wrap(f"\n\nLlamafile model: {repr(ag.llamafile_model)}",
                               width=80))
         elif ag.frontend == "vllm":
             commit_message += '\n'.join(
-                textwrap.wrap(f"\n\nVLLM model: {ag.vllm_model}.", width=80))
+                textwrap.wrap(f"\n\nVLLM model: {repr(ag.vllm_model)}", width=80))
     else:
         commit_message += "\n\n<Explain why change was made.>"
     commit_message += "\n\nNote, this commit message is generated by `debgpt git commit`."
@@ -223,10 +208,14 @@ def gather_information_ordered(msg: Optional[str], ag,
         if key == 'mapreduce':
             spec = ag.mapreduce.pop(0)
             aggregated = mapreduce.mapreduce_super_long_context(
-                spec, ag.mapreduce_chunksize, ag.frontend_instance, ag.ask,
-                ag.debgpt_home, ag.verbose,
+                spec,
+                ag.mapreduce_chunksize,
+                ag.frontend_instance,
+                ag.ask,
+                ag.verbose,
+                ag.mapreduce_map_mode == 'compact',
                 ag.mapreduce_reduce_mode == 'compact',
-                ag.mapreduce_parallelism)
+                parallelism=ag.mapreduce_parallelism)
             msg = _append_info(msg, aggregated)
         elif key == 'retrieve':
             raise NotImplementedError(key)
@@ -234,7 +223,7 @@ def gather_information_ordered(msg: Optional[str], ag,
             raise NotImplementedError(key)
         elif key in ('file', ):
             spec = getattr(ag, key).pop(0)
-            func = ft.partial(reader.read_and_wrap, debgpt_home=ag.debgpt_home)
+            func = ft.partial(reader.read_and_wrap)
             msg = _append_info(msg, func(spec))
         elif key == 'inplace':
             # This is a special case. It reads the file as does by
@@ -291,6 +280,18 @@ def _dispatch_subcommand(ag):
         pass
 
 
+def sideeffect_cache_refresh() -> None:
+    '''
+    handles the cache refresh, automatically delete the expired cache
+    entries. We need to do this because the cache expire seems not
+    thread safe and would cause trouble if we don't do it before
+    entering anything with concurrent.futures.ThreadPoolExecutor.
+    '''
+    # triggers automatic cache expire
+    c = cache.Cache(defaults.CACHE)
+    c.close()
+
+
 def sideeffect_output(ag: object, f: frontend.AbstractFrontend) -> None:
     '''
     handles the output specified by --output argument
@@ -360,6 +361,9 @@ def main(argv=sys.argv[1:]):
             os.path.expanduser('~/.debgpt/config.toml'))
         exit(0)
 
+    # refresh debgpt cache
+    sideeffect_cache_refresh()
+
     # process subcommands. Note, the subcommands will exit() when finished.
     # some subcommands will require a frontend instance, such as git commit.
     _dispatch_subcommand(ag)
@@ -387,26 +391,35 @@ def main(argv=sys.argv[1:]):
         console.print(msg, markup=False)
         exit(0)
 
-    # print the prompt and do the first query, if specified
-    if msg is not None:
-        if not ag.hide_first:
-            console.print(Panel(escape(msg), title='Initial Prompt'))
+    pending_exc = None
 
-        # query the backend
-        frontend.interact_once(f, msg)
+    try:
+        # print the prompt and do the first query, if specified
+        if msg is not None:
+            if not ag.hide_first:
+                console.print(Panel(escape(msg), title='Initial Prompt'))
 
-    # drop the user into interactive mode if specified (-i)
-    if not ag.quit:
-        frontend.interact_with(f)
+            # query the backend
+            frontend.interact_once(f, msg)
 
-    # inplace mode: write the LLM response back to the file
-    sideeffect_inplace(ag, f)
+        # drop the user into interactive mode if specified (-i)
+        if not ag.quit:
+            frontend.interact_with(f)
 
-    # let frontend dump session to json under debgpt_home
-    f.dump()
-    # handle the --output argument
-    sideeffect_output(ag, f)
+        # inplace mode: write the LLM response back to the file
+        sideeffect_inplace(ag, f)
 
+        # handle the --output argument
+        sideeffect_output(ag, f)
+    except Exception as e:
+        # re-raise the exception later
+        pending_exc = e
+    finally:
+        # let frontend dump session to json under debgpt_home
+        f.dump()
+
+    if pending_exc is not None:
+        raise pending_exc
 
 if __name__ == '__main__':
     main()
