@@ -26,13 +26,11 @@ import functools as ft
 import sys
 import glob
 import shlex
-import mimetypes
 import tenacity
 from rich.rule import Rule
 from rich.progress import track
 import concurrent.futures
 from urllib.parse import urlparse
-from urllib.request import urlopen
 import urllib.parse
 from . import policy as debian_policy
 from .defaults import console
@@ -63,6 +61,41 @@ HEADERS = {
 }
 
 
+def help():
+    '''
+    Show the list of supported reader specs.
+
+    Use any of the following commands to show the help message:
+        debgpt -f :
+        debgpt -f ?
+        debgpt -x :
+        debgpt -x ?
+    '''
+    console.print(Rule('Supported reader specs (specifiied by -f or -x)'))
+    console.print(' • <file_path>       read plain text file')
+    console.print(' • <directory>       read all files under a directory')
+    console.print(' • http://<...>      read plain text from http URL')
+    console.print(' • https://<...>     read plain text from https URL')
+    console.print(' • file://<...>      read plain text from file URL')
+    console.print(' • archwiki:<name>   read ArchWiki page')
+    console.print(' • bts:<bugnumber>   read Debian BTS bug report')
+    console.print(' • buildd:<pkg>      read Debian buildd status')
+    console.print(' • cmd:<cmd_line>    read the output of a command line')
+    console.print(' • devref:all        load whole Debian Devref, full text')
+    console.print(' • devref:<section>  load a specific section of Debian Devref')
+    console.print(' • google:<query>    load the topmost Google search results')
+    console.print(' • ldo:<list>        read the mailing list threads from lists.debian.org, e.g. "debian-ai/2024/11"')
+    console.print(' • man:<man>         read the manual page of a command')
+    console.print(' • policy:all        load whole Debian Policy, full text')
+    console.print(' • policy:<section>  load a specific section of Debian Policy')
+    console.print(' • sbuild:           load the latest sbuild buildlog, and filter out the unimportant lines')
+    console.print(' • sbuild:<path>     load the latest sbuild log from path, and filter out the unimportant lines')
+    console.print(' • tldr:<cmd>        read the tldr of a command')
+    console.print(' • stdin             read from stdin')
+    console.print(' • -                 read from stdin')
+    console.print(Rule('Please specify one or multiple of the above specs.'))
+
+
 def enable_cache(func: callable) -> callable:
     '''
     Enable caching for the function based on the first arg.
@@ -162,21 +195,57 @@ def read_file_plaintext(path: str) -> str:
     return content
 
 
-def read_sbuild(*, return_path: bool = False) -> str:
-    '''
-    load the latest sbuild buildlog. we will automatically figure out the
-    latest buildlog file in the parent directory.
-    '''
-    if not os.path.exists('./debian'):
-        raise FileNotFoundError(
-            './debian directory not found. Cannot detect sbuild log location. Are you in the right directory?'
-        )
-    latest_build_log = latest_glob('../*.build')
-    result = read_file_plaintext(latest_build_log)
-    if return_path:
-        return result, latest_build_log
+def extract_build_changes(text):
+  """
+  ! written by Gemini 1.5 Pro
+  Extracts the part of a string that starts with a Build header and
+  ends with a Changes header, using regular expressions, excluding the markers.
+
+  Args:
+    text: A string containing the text to extract from.
+
+  Returns:
+    A string representing the extracted part, or None if no match is found.
+
+  """
+  start_pattern = r"\+[-]+\+\n\| Build +\|\n\+[-]+\+"
+  end_pattern = r"\+[-]+\+\n\| Changes +\|\n\+[-]+\+"
+
+  try:
+    # Find the start and end positions using regex
+    start_match = re.search(start_pattern, text)
+    end_match = re.search(end_pattern, text)
+
+    if start_match:
+      if end_match:
+        return text[start_match.end() : end_match.start()]
+      else:
+        return text[start_match.end() :]  # Match to the end if end_match is None
     else:
-        return result
+      return None
+  except AttributeError:
+    return None
+
+
+def read_sbuild(path: Optional[str] = None, *, return_path: bool = False) -> str:
+    '''
+    load the latest sbuild buildlog. If no path is provided, we will
+    automatically figure out the latest buildlog file in the parent directory.
+    Additionally, we will filter the unimportant lines from the buildlog.
+    '''
+    if path is None:
+        if not os.path.exists('./debian'):
+            raise FileNotFoundError(
+                './debian directory not found. Cannot detect sbuild log location. Are you in the right directory?'
+            )
+        latest_build_log = latest_glob('../*.build')
+        result = read_file_plaintext(latest_build_log)
+    else:
+        latest_build_log = path
+        result = read_file_plaintext(path)
+    # filter out the unimportant lines
+    result = extract_build_changes(result)
+    return (result, latest_build_log) if return_path else result
 
 
 def read_file_pdf(path: str) -> str:
@@ -448,6 +517,39 @@ def read_bts(spec: str) -> str:
     return '\n'.join(text)
 
 
+def fetch_ldo_threads(spec: str) -> List[str]:
+    '''
+    read the mail threads (including the links) from the lists.debian.org.
+    Return a list of URLs to the emails.
+
+    Example URL:
+    https://lists.debian.org/debian-ai/2024/11/threads.html
+                             ^^^^^^^^^^^^^^^^^
+                             spec (Specifier)
+    '''
+    url = f'https://lists.debian.org/{spec}/threads.html'
+    response = requests.get(url)
+    if response.status_code != 200:
+        console.log(f'Failed to read {url}: HTTP {response.status_code}')
+        return list()
+    soup = BeautifulSoup(response.text, features='html.parser')
+    links = soup.find_all('a', href=re.compile(r'^msg.*'))
+    links = [x.get('href') for x in links]
+    urls = [f'https://lists.debian.org/{spec}/{x}' for x in links]
+    console.log(f'Got {len(urls)} threads from {url}.')
+    return urls
+
+
+def read_ldo_threads(spec: str) -> List[Tuple[str, str]]:
+    urls = fetch_ldo_threads(spec)
+    with concurrent.futures.ThreadPoolExecutor() as executor:
+        results = list(track(executor.map(read_url, urls),
+                             total=len(urls),
+                             transient=True,
+                             description='Reading l.d.o threads'))
+    return [(x, y) for x, y in zip(urls, results)]
+
+
 def read_stdin() -> str:
     lines = [x.rstrip() for x in sys.stdin.readlines()]
     return '\n'.join(lines)
@@ -523,15 +625,15 @@ def read_buildd(spec: str):  # pragma: no cover
 def read(spec: str,
          *,
          user_question: Optional[str] = None,
-         debgpt_home: str = '.') -> List[Entry]:
+         ) -> List[Entry]:
     '''
     Unified reader for reading text contents from various sources
     specified by the user. We will detect the type of the resource specified,
     and dispatch to the corresponding reader.
 
     Args:
-        spec: the path or URL to the file
-        debgpt_home: the home directory of debgpt
+        spec (str): the path or URL to the file
+        user_question (str): the user question
     Returns:
         List[Entry]: a list of tuples, each tuple contains the parsed spec and
         the content, and two wrapper functions to wrap the content with.
@@ -568,7 +670,10 @@ def read(spec: str,
 
     results: List[Tuple[str, str]] = []
     # standard cases: file, directory, URL
-    if os.path.exists(spec) and os.path.isfile(spec):
+    if spec in (':', '?'):
+        help()
+        raise SystemExit()
+    elif os.path.exists(spec) and os.path.isfile(spec):
         parsed_spec = spec
         content = read_file(spec)
         wrapfun = create_wrapper('Here is the contents of file `{}`:', spec)
@@ -670,6 +775,15 @@ def read(spec: str,
             wrapfun_chunk = create_chunk_wrapper(
                 'Here is the contents from URL `{}` (lines {}-{}):', url)
             results.append((url, content, wrapfun, wrapfun_chunk))
+    elif spec.startswith('ldo:') or spec.startswith('lists.debian.org:'):
+        parsed_spec = spec[4:] if spec.startswith('ldo:') else spec[18:]
+        pairs = read_ldo_threads(parsed_spec)
+        for url, content in pairs:
+            wrapfun = create_wrapper('Here is the contents from URL `{}`:',
+                                     url)
+            wrapfun_chunk = create_chunk_wrapper(
+                'Here is the contents from URL `{}` (lines {}-{}):', url)
+            results.append(Entry(url, content, wrapfun, wrapfun_chunk))
     elif spec.startswith('man:'):
         parsed_spec = spec[4:]
         content = read_cmd(f'man {parsed_spec}')
@@ -709,7 +823,11 @@ def read(spec: str,
                 section = content[sectionidx]
                 results.append((source, section, wrapfun, wrapfun_chunk))
     elif spec.startswith('sbuild:'):
-        content, logpath = read_sbuild(return_path=True)
+        if spec == 'sbuild:':
+            content, logpath = read_sbuild(return_path=True)
+        else:
+            path = spec[7:]
+            content, logpath = read_sbuild(path, return_path=True)
         wrapfun = create_wrapper('Here is the sbuild buildlog {}:', logpath)
         wrapfun_chunk = create_chunk_wrapper(
             'Here is the sbuild buildlog {} (lines {}-{}):', logpath)
@@ -847,7 +965,7 @@ def read_and_chunk(spec: str,
                    *,
                    max_chunk_size: int = -1,
                    user_question: Optional[str] = None,
-                   debgpt_home: str = '.') -> List[Entry]:
+                   ) -> List[Entry]:
     '''
     Read contents from the specified resource and chunk the content into pieces.
 
@@ -855,11 +973,10 @@ def read_and_chunk(spec: str,
         spec (str): the path or URL to the file
         max_chunk_size (int): the maximum chunk size of the content. If the
             number is less than 0, we shall not chunk the contents.
-        debgpt_home (str): the home directory of debgpt
     Returns:
         List[Entry]: a list of entries, each entry contains a chunk of the content
     '''
-    entries = read(spec, user_question=user_question, debgpt_home=debgpt_home)
+    entries = read(spec, user_question=user_question)
     if max_chunk_size > 0:
         entries = ft.reduce(list.__add__,
                             [chunk_entry(x, max_chunk_size) for x in entries])
@@ -870,7 +987,7 @@ def read_and_wrap(spec: str,
                   *,
                   max_chunk_size: int = -1,
                   user_question: Optional[str] = None,
-                  debgpt_home: str = '.') -> str:
+                  ) -> str:
     '''
     Read contents from the specified resource and wrap the content to make it
     suitable for prompting LLM.
@@ -879,11 +996,10 @@ def read_and_wrap(spec: str,
         spec (str): the path or URL to the file
         max_chunk_size (int): the maximum chunk size of the content. If the
             number is less than 0, we shall not chunk the contents.
-        debgpt_home (str): the home directory of debgpt
     Returns:
         str: the wrapped content
     '''
-    entries = read(spec, user_question=user_question, debgpt_home=debgpt_home)
+    entries = read(spec, user_question=user_question)
     if max_chunk_size > 0:
         entries = ft.reduce(list.__add__,
                             [chunk_entry(x, max_chunk_size) for x in entries])
@@ -915,18 +1031,12 @@ def main(argv: List[str] = sys.argv[1:]):
                         type=int,
                         default=-1,
                         help='chunk the content into pieces')
-    parser.add_argument('--debgpt_home',
-                        type=str,
-                        default='.',
-                        help='the home directory of debgpt')
     args = parser.parse_args(argv)
 
     tally = 0
     if args.wrap:
         for file in args.file:
-            string = read_and_wrap(file,
-                                   max_chunk_size=args.chunk,
-                                   debgpt_home=args.debgpt_home)
+            string = read_and_wrap(file, max_chunk_size=args.chunk)
             console.print(Rule())
             console.log('Specifier:', file)
             console.print(string)
@@ -934,7 +1044,7 @@ def main(argv: List[str] = sys.argv[1:]):
         console.print('Total number of texts:', tally)
     else:
         for file in args.file:
-            entries = read(file, debgpt_home=args.debgpt_home)
+            entries = read(file)
             if args.chunk > 0:
                 entries = ft.reduce(
                     list.__add__,
